/***********************************************************************
 *
 * FileName:
 *		PnpNet.c
 *
 * Description:
 *    Implementation of network access functions for "Palm Network
 *    Programming" sample application.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 *    All network error checking code has been retained for purposes of
 *    example.
 *
 * Copyright  2000, Greg Winton
 * All rights reserved.
 *
 **********************************************************************/

#include "PnpNet.h"     // PnpNetConnect(), PnpNetDisconnect().
#include "PnpLog.h"     // PnpLogAddCaption(), PnpLogAddText(),
                        //    PnpLogAddResources().
#include <NetMgr.h>     // NetLib API.

// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
void PrvNetLogOpenCount (void);
void PrvNetLogInterfaces (void);

// --------------------------------------------------------------------
// Private variable declarations
// --------------------------------------------------------------------
static UInt16 sNetLibRefNum = 0;            // NetLib reference number.
static UInt16 sOpenCount = 0;               // Local open count.
static UInt16 sState = pnpNetStateUninit;   // Network subsystem state.

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpNetState:
   Get the current state of the Network subsystem.
   Params:  (none)
   Returns: Current state of network subsystem.
*/
UInt16
PnpNetState (void)
{
   return sState;
}

/*
   PnpNetInit:
   Initialize the Network subsystem.  Checks for the
   existance and version of the net library, then
   obtains a reference number to it.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpNetInit ()
{
   // Declare local variables.
   Err err = 0;            // Error code.
   UInt32 version = 0L;    // NetLib version.
   UInt16 major = 0;       // Major version.
   UInt16 minor = 0;       // Minor version.
   Char text [20] = "";    // Message text.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Initializing Network Library");

   // Check the ROM version.
   FtrGet (sysFtrCreator, sysFtrNumROMVersion, &version);

   // Extract the major, minor ROM versions.
   major = sysGetROMVerMajor(version);
   minor = sysGetROMVerMinor(version);

   // Log ROM version.
   StrPrintF (text, "ROM version %i.%i", major, minor);
   PnpLogAddItem (pnpLogNet, text);
   
   // Get the NetLib version.
   err = FtrGet (netFtrCreator, netFtrNumVersion, &version);

   // Check if NetLib version retrieved.
   if (0 != err)
   {
   // NetLib not installed.

      // Show error.
      PnpLogAddItem (pnpLogNet, "NetLib not available.");

      // Return error.
      return err;
   }

   // Extract the major, minor NetLib versions.
   major = sysGetROMVerMajor(version);
   minor = sysGetROMVerMinor(version);

   // Tell user the NetLib version.
   StrPrintF (text, "NetLib version %i.%i", major, minor);
   PnpLogAddItem (pnpLogNet, text);

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib found");

   // Find the net library.
   err = SysLibFind ("Net.lib", &sNetLibRefNum);

   // Check if net library located.
   if (0 != err)
   {
   // Library not found.

      // Show error.
      PnpLogAddItem (pnpLogNet, "NetLib not found.");

      // Return error.
      return err;
   }

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "NetLib found");

   // Update state, we are now initialized.
   sState = pnpNetStateInit;

   // Return.
   return err;
}

/*
   PnpNetConnect:
   Connect to the network
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpNetConnect ()
{
   // Declare local variables.
   Err err = 0;         // Error code.
   UInt16 ifErrs = 0;     // Interface errors.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Connecting to Network");

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib open");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Open the Net Library.
   err = NetLibOpen (sNetLibRefNum, &ifErrs);

   // Check if no errors.
   if (!err)
   {
   // Open succeeded.

      // Show available resources.
      PnpLogAddResources (pnpLogNet, "NetLib open");
   }

   // Check if the error is that we're already open.
   else if (netErrAlreadyOpen == err)
   {
   // NetLib is already open.

      // Show available resources.
      PnpLogAddResources (pnpLogNet, "NetLib already open");
   }

   else
   {
   // Open failed.

      // Draw current state.
      PnpLogAddItem (pnpLogNet, "Error opening NetLib.");

      // Return the error.
      return err;
   }

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Check if any interface errors occurred.
   if (ifErrs)
   {
   // One or more interface errors occurred.

      // Inform user of interface errors.
      PnpLogAddItem (pnpLogNet, "One or more interface errors");
   }

   // Show the interfaces and their states.
   PrvNetLogInterfaces ();

   // Increment local open count.
   ++sOpenCount;

   // Update state, we are now connected.
   sState = pnpNetStateConnect;

   // Return success.
   return 0;
}

/*
   PnpNetDisconnect:
   Disconnect from the network
   Parms:   (none)
   Returns: 0        Success.
*/
Err
PnpNetDisconnect ()
{
   // Declare local variables.
   Err err = 0;         // Error code.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Disconnecting from Network");

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib close");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Close the Net Library.
   err = NetLibClose (sNetLibRefNum, false);

   // Check if library was not open.
   if (netErrNotOpen == err)
   {
   // NetLib is already open.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "NetLib was not open.");

      // Return.
      return 0;
   }

   // Check if we're still open.
   else if (netErrStillOpen == err)
   {
   // NetLib is already open.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "NetLib is still open.");
   }

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "NetLib close pending");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Now close the net lib.
   NetLibFinishCloseWait (sNetLibRefNum);

   // Draw current state.
   PnpLogAddResources (pnpLogNet, "NetLib closed");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Show the interfaces and their states.
   PrvNetLogInterfaces ();
   
   // Decrement local open count.
   --sOpenCount;

   // Return error.
   return 0;
}

/*
   PnpNetTerm:
   Terminate the Network subsystem.  Checks if the
   net library is still open.  If it is, closes it,
   but with a delay.
   Parms:   (none)
   Returns: 0        Success.
   Note:    This function contains no user interface
            code.
*/
Err
PnpNetTerm ()
{
   // Declare local variables.

   // Close any remaining opens.
   while (sOpenCount)
   {
      // Close net library.
      NetLibClose (sNetLibRefNum, false);

      // Decrement open count.
      --sOpenCount;
   }

   // Return success.
   return 0;
}

/*
   PrvNetLogOpenCount:
   Show the Net Library's open count.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
void
PrvNetLogOpenCount (void)
{
   // Declare local variables.
   UInt16 openCount = 0;
   Char* textP = NULL;

   // Allocate a block for the text message off the heap.
   textP = (Char*) MemPtrNew (20);

   // Check memory allocation.
   if (NULL == textP)
   {
   // Memory allocation failed.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "Memory Allocation Failed!");

      // Return error.
      return;
   }

   // Get the open count.
   NetLibOpenCount (sNetLibRefNum,     // Net Lib ref num.
                    &openCount);       // Target for count.

   // Format message.
   StrPrintF (textP, "Open Count: %i", openCount);

   // Show message.
   PnpLogAddItem (pnpLogNet, textP);

   // Free the memory.
   MemPtrFree (textP);

   // Return.
   return;
}

/*
   PrvNetLogInterfaces:
   Show the state of each attached interface.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
void
PrvNetLogInterfaces ()
{
   // Declare local variables.
   Err err = 0;            // Return value.
   UInt16 index = 0;       // Iterator.
   UInt32 ifCreator = 0L;  // Interface creator id.
   UInt16 ifInstance = 0;  // Interface instance.
   UInt8 ifUp = false;     // Interface up/down flag.
   Char ifName [32] = "";  // Interface name buffer.
   UInt16 settingSize = 0; // Size of requested setting.
   Char* textP = NULL;     // Output string.

   // Allocate a block for the text message off the heap.
   textP = (Char*) MemPtrNew (50);

   // Check memory allocation.
   if (NULL == textP)
   {
   // Memory allocation failed.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "Memory Allocation Failed!");

      // Return.
      return;
   }

   // Traverse list of interfaces.
   while (true)
   {
      // Get the interface.
      err = NetLibIFGet (sNetLibRefNum,   // NetLib ref num.
                         index,           // 0-based index of interface.
                         &ifCreator,      // Interface's creator.
                         &ifInstance);    // Interface's instance number.

      // Check for error.
      if (err)
      {
      // An error occurred.

         // Break out of for loop.
         break;
      }

      // Get the interface name.
      StrCopy (ifName, "<Unknown>");
      settingSize = sizeof (ifName);
      err = NetLibIFSettingGet (sNetLibRefNum,      // NetLib ref num.
                          ifCreator,          // Interface's creator.
                          ifInstance,         // Interface's instance number.
                          netIFSettingName,   // Setting to retrieve.
                          ifName,             // Target for setting.
                          &settingSize);      // Size of target for setting.

      // Check if interface is up.
      settingSize = sizeof (ifUp);
      err = NetLibIFSettingGet (sNetLibRefNum,   // NetLib ref num.
                          ifCreator,       // Interface's creator.
                          ifInstance,      // Interface's instance number.
                          netIFSettingUp,  // Setting to retrieve.
                          &ifUp,           // Target for setting.
                          &settingSize);   // Size of target for setting.

      // Format message.
      StrPrintF (textP, "%s - %s", ifName, ifUp ? "up" : "down");

      // Show message.
      PnpLogAddItem (pnpLogNet, textP);

      // Increment the index.
      index++;
   }

   // Free the memory.
   MemPtrFree (textP);

   // Update state, we are now disconnected.
   sState = pnpNetStateDisconnect;

   // Return.
   return;
}
