/***********************************************************************
 *
 * FileName:
 *		PnpNet.h
 *
 * Description:
 *    Declaration of network access functions for "Palm Network
 *    Programming" sample application.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

#ifndef __PNPNET_H__
#define __PNPNET_H__

// Standard application definitions
#include "PnpDefs.h"
#include "GData.h"   // GDataPtr.

// --------------------------------------------------------------------
// Public constant declarations
// --------------------------------------------------------------------

// Invalid socket constant.
#define kNilSocket ((NetSocketRef) -1)
#define kInetAddrStringSize   (4*3 + 3*1 + 1 + 4 + 1)

// Network state constants.
#define pnpNetStateUninit       0
#define pnpNetStateInit         1
#define pnpNetStateConnect      2
#define pnpNetStateDisconnect   3

// Buffer Sizes
#define kMaxDotDec      15
#define kMaxHost        255
#define kMaxServ        15
#define kMaxProto       3

// --------------------------------------------------------------------
// Public function declarations
// --------------------------------------------------------------------

/*
   PnpNetState:
   Get the current state of the Network subsystem.
   Params:  (none)
   Returns: Current state of network subsystem.
*/
UInt16 PnpNetState (void);

/*
   PnpNetInit:
   Initialize the Network subsystem.  Checks for the
   existance and version of the net library, then
   obtains a reference number to it.
   Params:  (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err PnpNetInit (void);

/*
   PnpNetConnect:
   Connect to the network
   Params:  (none)
   Returns: 0        Success.
            1        Success, but one or more interfaces failed.
            < 0      Error.
*/
Err PnpNetConnect (void);

/*
   PnpNetDisconnect:
   Disconnect from the network
   Params:  (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err PnpNetDisconnect (void);

/*
   PnpNetTerm:
   Terminate the Network subsystem.  Checks if the
   net library is still open.  If it is, closes it,
   but with a delay.
   Params:  (none)
   Returns: 0        Success.
   Note:    This function contains no user interface
            code.
*/
Err PnpNetTerm (void);

/*
   PnpNetFindName:
   Retrieve the ip address for a host name.
   Params:  hostNameP      Host name to lookup.
   Returns: > 0            Host ip address.
            0              Error finding host name.
*/
NetIPAddr PnpNetFindName (Char* hostNameP);

/*
   PnpNetFindAddress:
   Retrieve the ip address for a host name.
   Params:  hostAddrP      Host address to lookup.
   Returns: > 0            Host ip address.
            0              Error finding host name.
*/
NetIPAddr PnpNetFindAddress (Char* hostAddrP);

/*
   PnpNetFindService:
   Retrieve the port number for a standard TCP/IP service.
   Params:  protoNameP     Protocol to use (TCP or UDP).
            servNameP      Service to retrieve.
   Returns: > 0            Port number.
            0              Error finding port.
*/
UInt16 PnpNetFindService (Char* protoNameP, Char* servNameP);

/*
   PnpNetSocketConnect:
   Open a socket connection.
   Params:  hostNameP   Remote host name/address string.
            servNameP   Service name string.
            protoNameP  Protocol name string (TCP or UDP).
   Returns: >= 0        Success, return value is socket ref.
            -1          Error.
*/
NetSocketRef PnpNetSocketConnect (Char* hostNameP, Char* servNameP, Char* protoNameP);

/*
   PnpNetSocketDisconnect:
   Close a socket connection.
   Params:  socketRef   Socket reference.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketDisconnect (NetSocketRef socketRef);

/*
   PnpNetSocketShutdown:
   Shutdown a socket connection.
   Params:  socketRef   Socket reference.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketShutdown (NetSocketRef socketRef);

/*
   PnpNetSocketReceive:
   Read a block of data from a socket.
   Params:  socketRef   Socket reference.
            dataP       Target for received data.
            sizeP       [in] Size of target buffer.
                        [out] Amount of data received.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketReceive (NetSocketRef socketRef, UInt8* textP, UInt16* sizeP);

/*
   PnpNetSocketSendLine:
   Write a block of text to the socket.
   Params:  socketRef   Socket reference.
            dataP       Data to send.
            sizeP       [IN]  Amount of data to send.
                        [OUT] Amount of data sent.
   Returns: 0           Success.
            < 0         Error.
   Note:    This function appends a carriage-return
            line-feed to the text being written.
*/
Err PnpNetSocketSend (NetSocketRef socketRef, UInt8* dataP, UInt16* sizeP);

/*
   PnpNetSelect:
   Check for I/O on one or more sockets. This function auto-
   matically checks for user input events such as a pen tap
   or key press.
   Params:  width       Highest socket ref in readFDs or writeFDs.
            readFDs     Set of descriptors to check for input.
            writeFDs    Set of descriptors to check for output.
   Return:  0           Success.
            <0          Error occurred.
*/
Err PnpNetSelect (UInt16 width, NetFDSetType* readFDs, NetFDSetType* writeFDs);

/*
   PnpNetSocketListen:
   Put the socket into passive, listening mode.
   Params:  (none)
   Returns: >= 0        Success, return value is socket ref.
            kNilSocket  Error.
*/
NetSocketRef PnpNetSocketListen (void);

/*
   PnpNetSocketAccept:
   Put the socket into passive, listening mode.
   Params:  listenRef   Listening socket to accept connection from.
   Returns: >= 0        Success, return value is socket ref.
            kNilSocket  Error.
*/
NetSocketRef PnpNetSocketAccept (NetSocketRef listenRef);

/*
   PnpNetSocketDmReceive:
   Receive data directly at end database record.
   Params:  socketRef   Socket to read data from.
            dataP       Data object.
   Returns: 0           Success.
            < 0         Error. No data, record is full or other.
*/
Err PnpNetSocketDmReceive (NetSocketRef socketRef, GDataPtr dataP);

/*
   PnpNetSocketAddress:
   Get the socket's IP address in dotted decimal notation.
   Params:  socketRef   Socket reference.
            addrS       Target for address string.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketAddress (NetSocketRef socketRef, Char* addrS);

/*
   PnpNetSocketPort:
   Get the socket's port number.
   Params:  socketRef   Socket reference.
            portNumP    Target for number.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketPort (NetSocketRef socketRef, UInt16* portNumP);

#endif	// __PNPNET_H__
