/***********************************************************************
 *
 * FileName:
 *		PnpLog.c
 *
 * Description:
 *    Implementation of logging functions for "Palm Network Programming"
 *    FTP Client sample application.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Logging function declarations.
#include "PnpLog.h"
#include "PnpDefs.h"
#include "GLog.h"

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const Char* kLogName = "Log";
static const UInt16 kLogCard = 0;
static const UInt32 kLogCreator = PNP_FILE_CREATOR;
static const UInt32 kLogType = 'DATA';
static const UInt16 kLogMaxTextLen = 0xFF;
 
// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
static UInt32 PrvLogGetAvailMemory (void);

// --------------------------------------------------------------------
// Private data declarations
// --------------------------------------------------------------------
static Char sText [kLogMaxTextLen] = "";
static GLog sLog;

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpLogInit:
   Initialize logging subsystem.
   Params:  (none)
   Returns: 0           Success.
            <0          Error code.
*/
Err
PnpLogInit (void)
{
   // Declare local variables.
   Err err = 0;            // Return value.

   // Initialize the log.
   err = GLogInit (&sLog, PNP_FILE_CREATOR);

   // Return result.
   return err;
}

/*
   PnpLogTerm:
   Terminate logging subsystem.
   Params:  (none)
   Returns: 0           Success.
   Note:    This operation is currently a no-op.
*/
Err
PnpLogTerm (void)
{
   // Terminate the log.
   GLogTerm (&sLog);

   // Return success.
   return 0;
}

/*
   PnpLogNumItems:
   Get the number of items in a log.
   Parms:   logID       Log identifier.
   Returns: >=0         Number of items in log.
            <0          Error.
*/
Int16
PnpLogNumItems (UInt16 logID)
{
   // Declare local variables.
   Int16 numItems = 0;

   // Ask log for the items.
   numItems = GLogNumEntries (&sLog, logID);

   // Return value.
   return numItems;
}

/*
   PnpLogGetItem:
   Retrieve an item from the log.
   Params:  logID       Log identifier.
            text        Text to draw.
   Returns: 0           Success.
            <0          Error.
*/
Err
PnpLogGetItem (UInt16 logID, UInt16 itemIndex, Char* textP, UInt16* sizeP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Retrieve entry from log.
   err = GLogGetEntry (&sLog, logID, itemIndex, textP, sizeP);

   // Return error.
   return err;
}

/*
   PnpLogAddCaption:
   Add a caption to the log (bars above and below).
   Params:  logID       Id of log for caption.
            text        Caption text.
   Returns: 0           Success.
            <0          Error.
*/
Err
PnpLogAddCaption (UInt16 logID, Char* text)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   UInt16 textLen = StrLen (text);

   // Format static string with dashed lines.
   MemSet (sText, textLen, '-');

   // Make sure string is nil-terminated.
   sText [textLen] = '\0';

   // Add dashed line to log.
   err = PnpLogAddItem (logID, " ");

   if (!err)
   {
      // Add item to the log.
      err = PnpLogAddItem (logID, text);

      if (!err)
      {
         // Add dashed line to log.
         PnpLogAddItem (logID, sText);
      }
   }

   // Return error.
   return err;
}

/*
   PnpLogAddResources:
   Draw the current resources on the viewport.
   Params:  logID       Id of log for caption.
            text        Caption text.
   Returns: 0           Success.
            <0          Error.
*/
Err
PnpLogAddResources (UInt16 logID, Char* text)
{
   // Declare local variables.
   Err err = 0;               // Return value.
   UInt32 availMemory = 0;    // Available memory.

   // Get resources.
   availMemory = PrvLogGetAvailMemory ();

   // Round memory size to kbytes.
   availMemory = (availMemory + 500L) / 1000L;

   // Check if add'l text requested.
   if (text)
   {
      // Format memory for output.
      StrPrintF (sText,
                 "%liKB avail [%s]",
                 availMemory,
                 text);
   }
   else
   {
      // Format memory for output.
      StrPrintF (sText,
                 "%liKB avail",
                 availMemory);
   }

   // Add item to the log.
   PnpLogAddItem (logID, sText);
   
   // Return value.
   return err;
}

/*
   PnpLogAddItem:
   Add an item to the log.
   Params:  logID       Log identifier.
            text        Text to draw.
   Returns: 0           Success.
            <0          Error.
*/
Err
PnpLogAddItem (UInt16 logID, Char* text)
{
   // Declare local variables.
   Err err = 0;            // Return value.

   // Add entry to underlying GLog instance.
   GLogAddEntry (&sLog, logID, text);

   // Return value.
   return err;
}

/*
   PnpLogPurge:
   Purge all items from the log.
   Params:  logID    Category of log to purge.
   Returns: 0              Success.
            <0             Error.
*/
Err
PnpLogPurge (UInt16 logID)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Purge the underlying GLog instance.
   GLogPurge (&sLog, logID);

   return err;
}

static UInt32
PrvLogGetAvailMemory (void)
{
   // Declare local variables.
   UInt32 availMemory = 0;    // Available memory.
   UInt16 numHeaps = 0;       // Number of heaps to traverse.
   UInt16 i = 0;              // Iterator.
   UInt16 heapID;             // Identifier for each heap.
   UInt32 maxChunk = 0;       // Max chunk size (per heap).
   UInt32 freeMemory = 0;     // Free memory (per heap).

   // Get the number of heaps.
   numHeaps = MemNumHeaps (0);

   // Traverse the heaps.
   for (i = 0; i < numHeaps; i++)
   {
   // For each heap...
 
      // Get the heap identifier.
      heapID = MemHeapID (0, i);

      // Check if the heap is dynamic.
      if (MemHeapDynamic (heapID))
      {
      // Heap is dynamic.

         // Get amount of free memory (and size of max chunk).
         MemHeapFreeBytes (heapID, &freeMemory, &maxChunk);

         // Add free memory
         availMemory += freeMemory;
      }
   }

   // Return value.
   return availMemory;
}
