#include "GBuf.h"    // GBuf... declarations
#include <unix_stdarg.h>

// --------------------------------------------------------------------
// Private structure declarations
// --------------------------------------------------------------------
typedef struct GBufType
{
   UInt16   allocSize;     // Size of structure and buffer.
   UInt16   lockCount;     // Count of all locks.
   UInt16   readLockCount; // Number of read locks on the buffer.
   UInt16   blockSize;     // Size of buffer.
   UInt16   blockFull;     // Amount of buffer that has data.
   UInt8    block[1];      // Pointer to the data.
} GBufType, * GBufPtr;

// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
static GBufPtr PrvBufLock (GBufHandle bufH);
static void    PrvBufUnlock (GBufPtr bufP);
static void    PrvBufClear (GBufPtr bufP, UInt16 startAt, UInt16 clearLen);

// --------------------------------------------------------------------
// Public function implementations.
// --------------------------------------------------------------------

/*
   GBufNew:
   Create a new buffer object.
   Params:  sizeP       [In]  requested size for buffer.
                        [Out] actual size for buffer.
   Returns: !Null       Pointer to created buffer.
            Null        Error creating buffer.
*/
GBufHandle GBufNew (UInt16 size)
{
   // Initialize local variables.
   GBufHandle bufH = NULL;    // Return value. Buffer instance handle.
   GBufPtr bufP = NULL;       // Pointer to buffer instance.
   UInt16 allocSize = 0;      // Total allocation needed.
   
   // For now, make sure size is > 0.
   // (In future, use default if *sizeP == 0).
   ErrFatalDisplayIf (0 == size, "Zero size passed to GBufCreate.");
   
   // Calculate how big we need to make it.
   allocSize = sizeof (GBufType) + (size);
   
   // Now allocate the block.
   bufH = MemHandleNew (allocSize);
   
   // Allocation must succeed.
   ErrFatalDisplayIf (NULL == bufH, "Unable to allocate object in GBufCreate");

   // Lock the buffer instance.
   bufP = PrvBufLock (bufH);
   
   // Assign alloc size value.
   bufP->allocSize = MemPtrSize (bufP);
   
   // Assign block size value.
   bufP->blockSize = bufP->allocSize - sizeof (GBufType);
   
   // Assign full value (currently empty).
   bufP->blockFull = 0;
   
   // Clear buffer.
   PrvBufClear (bufP, 0, bufP->blockSize);

   // Unlock the buffer instance.
   PrvBufUnlock (bufP);
   
   // bufP is no longer valid.
   bufP = NULL;
   
   // Return value.
   return bufH;
}

/*
   GBufFree:
   Destroy a buffer object.
   Params:  bufH        Buffer instance to free.
   Returns: (none)
*/
void
GBufFree (GBufHandle bufH)
{
   // Release the buffer.
   MemHandleFree (bufH);
}
   
   
/*
   GBufSize:
   Get the size of a buffer.
   Params:  bufH     GBuffer instance.
   Returns: >=0         Size of buffer.
            -1          Error.
*/
UInt16
GBufSize (GBufHandle bufH)
{
   // Declare local variables.
   UInt16 size = 0;
   GBufPtr bufP = PrvBufLock (bufH);

   // Retrieve size.
   size = bufP->blockSize;
   
   // Unlock buffer.
   PrvBufUnlock (bufP);
   
   // Buffer is no longer valid.
   bufP = NULL;
   
   // Return value.
   return size;
}

/*
   GBufAvail:
   Get the available space of a buffer.
   Params:  bufH     GBuffer instance.
   Returns: >= 0        Available space in buffer.
            -1          Error.
*/
UInt16
GBufAvail (GBufHandle bufH)
{
   // Declare local variables.
   UInt16 size = 0;
   GBufPtr bufP = PrvBufLock (bufH);

   // Retrieve size.
   size = bufP->blockSize - bufP->blockFull;
   
   // Unlock buffer.
   PrvBufUnlock (bufP);
   
   // Buffer is no longer valid.
   bufP = NULL;
   
   // Return value.
   return size;
}

/*
   GBufEmpty:
   Empty all data in the buffer.
   Params:  bufH        Buffer instance.
   Returns: 0           Success.
*/
Err GBufEmpty (GBufHandle bufH)
{
   // Declare local variables.
   UInt16 size = 0;
   GBufPtr bufP = PrvBufLock (bufH);

   // Clear the entire buffer.
   PrvBufClear (bufP, 0, bufP->blockSize);

   // Set "full" amount to 0.
   bufP->blockFull = 0;
   
   // Unlock buffer.
   PrvBufUnlock (bufP);
   bufP = NULL;         // Buffer is no longer valid.
   
   // Return success.
   return 0;
}

/*
   GBufAppend
   Append a block of data to the end of the buffer.
   Params:  bufH     GBuffer instance.
            blockP      Block of data to append.
            size        Size of data block to append.
   Returns: >=0         Number of bytes appended.
            -1          Error.
*/
UInt16
GBufAppend (GBufHandle bufH, UInt8* blockP, UInt16 size)
{
   // Declare local variables.
   GBufPtr bufP = PrvBufLock (bufH);
   UInt8* appendP = bufP->block + bufP->blockFull;
   UInt16 appendSize = bufP->blockSize - bufP->blockFull;

   // Check how much we want to append.
   if (size < appendSize)
   {
   // Size is less then appendSize.
   
      // Update appendSize.
      appendSize = size;
   }
   
   // Now, append to buffer.
   MemMove (bufP->block + bufP->blockFull, blockP, appendSize);
   
   // Update block full count.
   bufP->blockFull += appendSize;
   
   // Unlock buffer.
   PrvBufUnlock (bufP);

   // Buffer is no longer valid.
   bufP = NULL;
   
   // Return success.
   return appendSize;
}

/*
   GBufAppendF
   Format and append a block of data to the end of the buffer.
   Params:  bufH        GBuffer instance.
            formatP     Format string (as used by StrPrintF).
            ...         Variable argument list.
   Returns: >=0         Number of bytes appended.
            -1          Error.
*/
UInt16
GBufAppendF (GBufHandle bufH, Char* formatP, ...)
{
   // Declare local variables.
   GBufPtr bufP = PrvBufLock (bufH);
   Char* appendP = (Char*) (bufP->block + bufP->blockFull);
   UInt16 appendSize = bufP->blockSize - bufP->blockFull;
   va_list argList;

   // Get the argument list.
   argList = va_start (argList, formatP);
   
   // Pass argument list to StrVPrintF.
   appendSize = StrVPrintF (appendP, formatP, argList);
   
   // Release argument list.
   va_end (argList);
   
   // Make sure we didn't overwrite the buffer.
   ErrFatalDisplayIf (bufP->blockFull + appendSize > bufP->blockSize,
                     "GBufAppendF has written past the end of a memory block.");
                     
   // Update full size.
   bufP->blockFull += appendSize;
   
   // Unlock buffer.
   PrvBufUnlock (bufP);

   // Buffer is no longer valid.
   bufP = NULL;
   
   // Return success.
   return appendSize;
}

/*
   GBufRead:
   Read a block of data from the start of the buffer.
   Params:  bufH     GBuffer instance.
            blockP      Out-param for read data, NULL just clears buffer.
            size        Max size to read.
   Returns: >=0         Number of bytes read/discarded.
            -1          Error.
*/
UInt16
GBufRead (GBufHandle bufH, UInt8* blockP, UInt16 size)
{
   // Declare local variables.
   GBufPtr bufP = PrvBufLock (bufH);
   UInt8* readP = bufP->block;
   UInt16 readSize = bufP->blockFull;
   UInt16 moveSize = 0;

   // Check if there is anything to read.
   if (readSize)
   {
   // There is something to read.

      // Check how much we want to read.
      if (size < readSize)
      {
      // Size is less then readSize.
      
         // Update readSize.
         readSize = size;
      }
      
      // See if caller wants the data.
      if (NULL != blockP)
      {
      // Caller wants data.
      
         // Read data to output buffer.
         MemMove (blockP, bufP->block, readSize);
      }
      
      // Calculate amount of data to move.
      moveSize = bufP->blockFull - readSize;
      
      // Check if there is data to move.
      if (0 < moveSize)
      {
      // There is data to move.
      
         // Move the data forward in the buffer.
         MemMove (bufP->block, bufP->block + readSize, bufP->blockFull - readSize);
      }
      
      // Clear the old buffer.
      PrvBufClear (bufP, moveSize, readSize);
      
      // Update full value.
      bufP->blockFull -= readSize;
   }
   else
   {
   // Nothing to read.
   
      // Set return value accordingly.
      readSize = -1;
   }
      
   // Unlock buffer.
   PrvBufUnlock (bufP);

   // Buffer is no longer valid.
   bufP = NULL;
   
   // Return success.
   return readSize;
}

/*
   GBufGetReadPtr:
   Get a read pointer to the current start of the buffer.
   Params:  bufH     GBuffer instance.
            sizeP       [In]  requested size in bytes.
                        [Out] size of buffer that can be read.
   Returns: !Null       Address of start of buffer.
            Null        An error occurred.
   Note:    The read pointer is available to only multiple owners. 
            Please be sure to release it when you are done.
*/
const UInt8* GBufGetReadPtr (GBufHandle bufH, UInt16* sizeP)
{
   // Declare local variables.
   GBufPtr bufP = PrvBufLock (bufH);
   UInt8* readP = bufP->block;
   UInt16 readSize = bufP->blockFull;

   // Check if there is anything to read.
   if (!readSize)
   {
   // Nothing to read.
   
      // Set read pointer accordingly.
      readP = NULL;
   }
   
   // Set the out-param.
   (*sizeP) = readSize;

   // Return pointer.
   return readP;
}

/*
   GBufReleaseReadPtr:
   Release the read pointer returned by GBufGetReadPtr.
   Params:  bufH        Buffer instance.
   Returns: 0           Success.
            -1          Error.
*/
Err
GBufReleaseReadPtr (GBufHandle bufH)
{
   // Declare local variables.
   GBufPtr bufP = PrvBufLock (bufH);

   // Release a lock for the original.
   PrvBufUnlock (bufP);
   
   // And release once for now. We'll crash if its wrong.
   PrvBufUnlock (bufP);
      
   // Return success.
   return 0;
}

// --------------------------------------------------------------------
// Private function implementations.
// --------------------------------------------------------------------

static GBufPtr
PrvBufLock (GBufHandle bufH)
{
   // Declare local variables.
   GBufPtr bufP = NULL;
   
   // Make sure buffer is valid.
   ErrFatalDisplayIf (NULL == bufH, "Null handle passed to PrvBufLock.");
   
   // Lock the buffer.
   bufP = (GBufPtr) MemHandleLock (bufH);
   
   // Increment lock count.
   ++bufP->lockCount;
   
   // Make sure lock succeeded.
   ErrFatalDisplayIf (NULL == bufP, "Unable to lock buffer handle in PrvBufLock.");
   
   // Return value.
   return bufP;
}

static void
PrvBufUnlock (GBufPtr bufP)
{
   // Make sure buffer is valid.
   ErrFatalDisplayIf (NULL == bufP, "Null pointer passed to PrvBufUnlock");
   
   // Decrement lock count.
   --bufP->lockCount;
   
   // Unlock the buffer.
   MemPtrUnlock (bufP);
}
   
static void
PrvBufClear (GBufPtr bufP, UInt16 startAt, UInt16 clearLen)
{
   // Make sure lengths make sense.
   ErrFatalDisplayIf (bufP->blockSize < startAt + clearLen,
                     "Invalid parameters passed to PrvBufClear.");
                     
   MemSet (bufP->block + startAt, clearLen, '\0');
}
