/***********************************************************************
 *
 * FileName:
 *		GLog.c
 *
 * Description:
 *    Implementation of GLib Log support.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Logging function declarations.
#include "GLog.h"
#include <DataMgr.h>

// --------------------------------------------------------------------
// Private structure declarations
// --------------------------------------------------------------------
typedef struct
{
   UInt32 time;
   UInt16 textLen;
   Char  text [1];  
} GLogEntryType, * GLogEntryPtr;

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const UInt16 kLogCard = 0;
static const UInt32 kLogType = 'GLOG';
static const Char* kLogPrefix = "GLog";
static const UInt16 kMaxLogName = sizeof kLogPrefix + 4;
 
// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------

/*
   PrvCalcDatabaseName:
   Calculate the database name for a given creator.
   Params:  nameP          Pointer to buffer for calculated name.
            creatorID      Creator ID of client application - encoded in name.
   Returns: 0              Success.
   Note:    This function assumes that the buffer pointed to by nameP
            is at least kMaxLogName bytes.
   Note:    Log names are formatted: "GLogXXXX" where XXXX is the creator id.
*/
static Err PrvCalcDatabaseName (Char* nameP, UInt32 creatorID);

/*
   PrvGetLogEntry:
   Retrieve an entry record.
   Params:  logP        Pointer to log object.
            logID       ID of log.
            entryIndex  Index of entry to retrieve.
   Returns: !NULL       Success.
            NULL        Error.
*/
static GLogEntryPtr PrvGetLogEntry (GLog* logP, UInt16 logID, UInt16 entryIndex);

/*
   PrvReleaseLogEntry:
   Release an entry record.
   Params:  entryP      Pointer to record entry.
   Returns: (none)
*/
static void PrvReleaseLogEntry (GLogEntryPtr entryP);
            
// --------------------------------------------------------------------
// Private data declarations
// --------------------------------------------------------------------

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   GLogInit:
   Initialize logging session.
   Params:  logP           GLog instance.
            creatorID      Creator ID for log client.
   Returns: 0              Success.
            glibErrFile    Unable to access log file.
*/
Err
GLogInit (GLogPtr logP, UInt32 creatorID)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   Char logName [kMaxLogName];
   LocalID localID = 0;
   
   // Calculate log name.
   err = PrvCalcDatabaseName (logName, creatorID);

   // Look for database.
   localID = DmFindDatabase (kLogCard, logName);

   // Check if database was found.
   if (0 == localID)
   {
   // Database not found.

      // Create log database.
      err = DmCreateDatabase (kLogCard, logName, creatorID, kLogType, false);
   }

   // Check for error.
   if (!err)
   {
   // Success to here.

      // Try to open database.
      logP->dbR = DmOpenDatabaseByTypeCreator (kLogType,
                                               creatorID,
                                               dmModeReadWrite);

      // Check if opened successfully.
      if (logP->dbR)
      {
         // Initialize rest of log structure.
         logP->creatorID = creatorID;
      }
      else
      {
      // Unable to open log file.

         // Set error code.
         err = glibErrFile;
      }
   }

   // Return result.
   return err;
}

/*
   GLogTerm:
   Terminate logging subsystem.
   Params:  logP     GLog instance.
   Returns: 0        Success.
            <0       Error code.
*/
Err
GLogTerm (GLogPtr logP)
{
   // Close the log database.
   DmCloseDatabase (logP->dbR);
   logP->dbR = NULL;

   // Clear log attributes.
   logP->creatorID = 0;

   // Return success.
   return 0;
}

/*
   GLogNumEntries:
   Get the number of entries in a log.
   Params:  logP     GLog instance.
            logID    Index of log.
   Returns: >=0      Number of entries in log.
            <0       Error.
*/
Int16
GLogNumEntries (GLogPtr logP, UInt16 logID)
{
   // Declare local variables.
   Int16 numEntries = 0;   // Return value.

   // Assume database is open.
   ErrFatalDisplayIf (!logP->dbR, "Log database is not open.");

   // Get the number of items in the category.
   numEntries = DmNumRecordsInCategory (logP->dbR, logID);

   return numEntries;
}

/*
   GLogGetEntry:
   Retrieve an entry from the log.
   Params:  logP           GLog instance.
            logID          Index of log.
            entryIndex      Index of entry in log.
            textP          Out-param for log entry.
            sizeP          [in]  Max size of textP.
                           [out] Number of characters copied to textP.
   Returns: 0              Success.
            <0             Error.
*/
Err
GLogGetEntry (GLogPtr logP, UInt16 logID, UInt16 entryIndex, Char* textP, UInt16* sizeP)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   GLogEntryPtr entryP = PrvGetLogEntry (logP, logID, entryIndex);

   // Make sure we got the entry.
   if (entryP)
   {
   // We got the entry.
   
      // Fill in out-params.
         
      // Check if we want to actually retrieve the text.
      if (textP)
      {
      // We want the text.

         // Clear out-param.
         MemSet (textP, (*sizeP), '\0');

         // Copy text to out-param.
         StrNCopy (textP, entryP->text, (*sizeP)-1);

         // Set size out-param.
         (*sizeP) = StrLen (textP) + 1;
      }
      else
      {
      // Don't want the text, only size.

         // Set size out-param.
         (*sizeP) = entryP->textLen + 1;
      }

      // Release the entry.
      PrvReleaseLogEntry (entryP);
   }
   else
   {
   // Record not found.

      // Set error to indicate record not found.
      err = glibErrNoData;
   }

   // Return value.
   return err;
}

/*
   GLogGetEntryTime:
   Retrieve the time an entry was written to the log.
   Params:  logP            GLog Instance.
            logID           ID of log.
            entryIndex      Index of log entry.
            timeP           Out-param for time of log entry.
   Returns: 0               Success.
            !0              Error.
*/
Err
GLogGetEntryTime (GLogPtr logP, UInt16 logID, UInt16 entryIndex, UInt32* timeP)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   GLogEntryPtr entryP = PrvGetLogEntry (logP, logID, entryIndex);

   // Make sure we got the entry.
   if (entryP)
   {
   // We got the entry.
   
      // Fill in out-params.
         
      // Set time out-param.
      (*timeP) = entryP->time;

      // Release the entry.
      PrvReleaseLogEntry (entryP);
   }
   else
   {
   // Record not found.

      // Set error to indicate record not found.
      err = glibErrNoData;
   }

   // Return value.
   return err;
}
                          
/*
   GLogAddEntry:
   Add an entry to the log.
   Params:  logID       Log identifier.
            text        Text to draw.
   Returns: 0           Success.
            <0          Error.
*/
Err
GLogAddEntry (GLogPtr logP, UInt16 logID, Char* text)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   MemHandle recH;
   GLogEntryPtr entryP;
   UInt16 pos, size, attr, textLen;
   void* recP = NULL;
   
   // Assume database is open.
   ErrFatalDisplayIf (!logP->dbR, "Log database is not open.");

   // Calc. length of text.
   textLen = StrLen (text);
   
   // Calc. record size.
   size = sizeof (*entryP) - sizeof (entryP->text) + textLen + 1;

   // Allocate a block of memory big enough for entry.
   entryP = (GLogEntryPtr) MemPtrNew (size);
   ErrFatalDisplayIf (!entryP, "Memory Allocation Failed");

   // Populate the structure.
   entryP->time = TimGetSeconds ();
   entryP->textLen = textLen;
   StrNCopy (entryP->text, text, textLen);
   entryP->text [textLen] = '\0';
   
   // Now we need a record for this entry.
   
   // Calc. record position - after last record. 
   pos = DmNumRecords (logP->dbR) + 1;

   // Add record to end of the database.
   recH = DmNewRecord (logP->dbR, &pos, size);
   ErrFatalDisplayIf (!recH, "Database Record Allocation Failed");

   // Lock the record.
   recP = MemHandleLock (recH);

   // Set the values.
   DmWrite (recP, 0, entryP, size);
   
   // Free the temp record.
   MemPtrFree (entryP);
   entryP = NULL;
   
   // Unlock the record.
   MemPtrUnlock (recP);
   recP = NULL;
      
   // Set record category.

   // Get record attribute word.
	DmRecordInfo (logP->dbR, pos, &attr, NULL, NULL);

   // Remove all category bits.
	attr &= ~dmRecAttrCategoryMask;

	// Set the attribute bits to indicate the current category.
	attr |= logID;

   // Set the record attrbitute word.
	err = DmSetRecordInfo (logP->dbR, pos, &attr, NULL);

   // Release record.
   DmReleaseRecord (logP->dbR, pos, true);

   // Return error.
   return err;
}

/*
   GLogPurge:
   Purge all items from the log.
   Params:  logP     GLog instance
            logID    Category of log to purge.
   Returns: 0        Success.
            <0       Error.
*/
Err
GLogPurge (GLogPtr logP, UInt16 logID)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Assume database is open.
   ErrFatalDisplayIf (!logP->dbR, "Log database is not open.");

   // Delete all records in category.
   err = DmDeleteCategory (logP->dbR, logID);

   // Return error.
   return err;
}

/*
   PrvCalcDatabaseName:
   Calculate the database name for a given creator.
   Params:  nameP          Pointer to buffer for calculated name.
            creatorID      Creator ID of client application - encoded in name.
   Returns: 0              Success.
   Note:    This function assumes that the buffer pointed to by nameP
            is at least kMaxLogName bytes.
   Note:    Log names are formatted: "LogXXXX" where XXXX is the creator id.
*/
static Err
PrvCalcDatabaseName (Char* nameP, UInt32 creatorID)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   Char* byteP = (Char*) &creatorID;

   // Copy the prefix.
   StrCopy (nameP, kLogPrefix);
   StrNCopy (nameP + StrLen (kLogPrefix), (Char*) &creatorID, 4);
   nameP [kMaxLogName-1] = '\0';
   
   // Return error.
   return err;
}

/*
   PrvGetLogEntry:
   Retrieve an entry record.
   Params:  logP        Pointer to log object.
            logID       ID of log.
            entryIndex  Index of entry to retrieve.
   Returns: !NULL       Success.
            NULL        Error.
*/
static GLogEntryPtr
PrvGetLogEntry (GLog* logP, UInt16 logID, UInt16 entryIndex)
{
   // Declare local variables.
   GLogEntryPtr eventP = NULL;  // Return value.
   UInt16 recIndex = 0;
   MemHandle recH = NULL;

   // Assume database is open.
   ErrFatalDisplayIf (!logP->dbR, "Log database is not open.");

   // Look for the record.
   if (0 == DmSeekRecordInCategory (logP->dbR,
                                    &recIndex,
                                    entryIndex,
                                    dmSeekForward,
                                    logID))
   {
      recH = DmQueryRecord (logP->dbR, recIndex);
      
      if (recH)
      {
         eventP = (GLogEntryPtr) MemHandleLock (recH);
      }
   }
   
   // Return value.
   return eventP;
}

/*
   PrvReleaseLogEntry:
   Release an entry record.
   Params:  entryP      Pointer to record entry.
   Returns: (none)
*/
static void
PrvReleaseLogEntry (GLogEntryPtr entryP)
{
   MemPtrUnlock (entryP);
}
            
